###### Assessing uncertainty in linked linear mixed models #########
##
## Sven Hohenstein, Hannes Matuschek
##
## Apr/May/Jun 2016

## Part 1: functions
library(lme4)

sampleFromModelFast <- function(object, size) {
  stopifnot(size >= 1, isLMM(object))
  
  ## Extract coefficients of the original model
  # means of the fixed effects
  fix_mean <- fixef(object)
  # covariance matrix of the fixed effects
  fix_vcov <- vcov(object)
  # number of fixed effects
  n_fix <- getME(object, "p")
  # model matrix (for the fixed effects)
  fix_moma <- model.matrix(object)
  # dependent variable
  DV <- getME(object, "y")
  # data set
  dat <- object@frame
  pred0 = predict(object, type="response")
  predresid0 = residuals(object, type="response")
  
  replicate(size, {
    ## Sample new deviation from est. fixed effects
    eta <- fix_moma %*% t(chol(fix_vcov)) %*% rnorm(n_fix)
    list(predictions = as.vector(pred0+eta), residuals = as.vector(predresid0-eta))
  }, simplify = FALSE)
}


sampleFromModel <- function(object, size) {
  stopifnot(size >= 1, isLMM(object))
  
  ## Extract coefficients of the original model
  # means of the fixed effects
  fix_mean <- fixef(object)
  # covariance matrix of the fixed effects
  fix_vcov <- vcov(object)
  # number of fixed effects
  n_fix <- getME(object, "p")
  # model matrix (for the fixed effects)
  fix_moma <- model.matrix(object)
  # dependent variable
  DV <- getME(object, "y")
  # data set
  dat <- object@frame
  # formula
  form <- formula(object)
  # name of the reponse variable
  resp <- deparse(form[[2L]])
  resp_sym <- as.symbol(resp)
  
  ## create minimal optimizer object
  OPT <- list(par = getME(object, "theta"),
              conv = object@optinfo$conv$opt,
              fval = object@devcomp$cmp[["dev"]])
  
  
  replicate(size, {
    ## Sample new fixed effects
    alpha <- fix_mean + t(chol(fix_vcov)) %*% rnorm(n_fix)
    
    ## formula for pure RE model
    form_no_fix <- reformulate(paste0("(", 
                                      vapply(findbars(form),
                                             lme4:::safeDeparse, 
                                             ""),
                                      ")"),
                               response = resp_sym, intercept = FALSE)
    
    ## remove "new" fixed effects from data
    robs <- as.vector(DV - ( fix_moma %*% alpha ))
    # replace old reponse vector with new one
    dat[[resp]] <- robs
    
    ## construct & "fit" pure RE model
    lrmod <- lFormula(form_no_fix, data = dat, REML = isREML(object))
    rdevfun <- do.call(mkLmerDevfun, lrmod)
    
    ## here we reuse the theta of the original model through "opt"
    rmod <- mkMerMod(environment(rdevfun), OPT, lrmod$reTrms, fr = lrmod$fr)
    
    ## Get predictions and residuals
    # get new predictions (fixed + re part)
    pred <- fix_moma %*% alpha + fitted(rmod) 
    # get new residuals
    res <- residuals(rmod)
    list(predictions = as.vector(pred), residuals = unname(res))
  }, simplify = FALSE)
}

# fit model based on predictions and residuals
fitLinkedModel <- function(pred, res, formula, data, REML) {
  stopifnot("pred__" %in% all.vars(formula) || "res__" %in% all.vars(formula))
  dat_ext <- cbind(data, "pred__" = pred, "res__" = res)
  lmer(formula, dat_ext, REML)
}

# reduce model size to save harddisk space
reduceModel <- function(model) {
  model@frame <- data.frame()
  model@flist <- list()
  model
}  



## Part 2: Models, Parameters
# base model
path1 <- "../../m_rfl_0_zcp.rds"
m_rfl_0_zcp <- readRDS(path1)

# linked model
path2 <- "../../m_fd_0_zcp_rfl_pred_res.rds"
m_fd_0_zcp_rfl_pred_res <- readRDS(path2)

# formula for new linked models
form_linked <- fd ~ poly(pred__, 2L) +
  poly(res__, 2L) + 
  l1.c + p1.c + f1.c +
  l.c + p.c + f.c + 
  l2.c + p2.c + f2.c + 
  (1 +
     l1.c + p1.c + f1.c +
     l.c + p.c + f.c + 
     l2.c + p2.c + f2.c || Subj) + 
  (1 | Word) + (1 | Sent)

# number of simulations
n <- 1000L

# output path
out_path <- "sim_res"

# output filename schema
out_filename <- "linked_sim_.rds"



## Part 3: Simulation loop
seed <- 9458
for (i in 1:n) {
  cat("Round:", i, "\n")
  # set seed
  set.seed(seed + i)
  # sample predictions and residuals
  cat("sampling...\n")
  pred_res_list <- sampleFromModelFast(m_rfl_0_zcp, 1L)[[1L]]
  pred <- pred_res_list[["predictions"]]
  res <- pred_res_list[["residuals"]]
  # fit model based on predictions and residuals
  cat("fitting linked model...\n")
  linked_mod <- fitLinkedModel(pred, res, form_linked,
                               m_fd_0_zcp_rfl_pred_res@frame,
                               isREML(m_fd_0_zcp_rfl_pred_res))
  # check convergence
  conv <- linked_mod@optinfo$conv$opt == 0L
  
  # summary
  sum_mod <- coef(summary(linked_mod))
  
  # reduce model
  linked_mod_red <- reduceModel(linked_mod)
  # save results
  cat("saving results...\n")
  res <- list(model = linked_mod_red, conv = conv, summary = sum_mod,
              predictions = pred, residuals = res)
  filename <- sub("(?=\\.rds$)", i, out_filename, perl = TRUE)
  saveRDS(res, file.path(out_path, filename))
}
